from typing import Optional, Callable

import panel
from atap_corpus.corpus.corpus import DataFrameCorpus
from panel.viewable import Viewer

from atap_corpus_loader.controller import Controller
from atap_corpus_loader.view import ViewWrapperWidget

panel.extension(notifications=True)


class CorpusLoader(Viewer):
    """
    Public interface for the CorpusLoader module. Maintains a reference to the logic Controller and the GUI wrapper.
    A CorpusLoader object can be used as a Panel component, i.e. will render in a Panel GUI.
    The build_callback_fn will be called when a corpus is built (can be set using set_build_callback()).
    """

    def __init__(self, root_directory: str, **params):
        """
        :param root_directory: The root directory that the file selector will search for files to load. The argument must be a string. The directory may be non-existent at initialisation time, but no files will be displayed until it exists.
        :param params: passed onto the panel.viewable.Viewer super-class
        :type root_directory: str
        """
        super().__init__(**params)
        self.controller: Controller = Controller(root_directory)
        self.view: ViewWrapperWidget = ViewWrapperWidget(self.controller)

    def __panel__(self):
        return self.view

    def set_build_callback(self, callback: Callable, *args, **kwargs):
        """
        Allows a callback function to be set when a corpus has completed building
        :param callback: the function to call when a corpus has been built
        :param args: positional arguments to pass onto the callback function
        :param kwargs: keyword arguments to pass onto the callback function
        :type callback: Callable
        :type args: Any
        :type kwargs: Any
        """
        self.controller.set_build_callback(callback, *args, **kwargs)

    def get_latest_corpus(self) -> Optional[DataFrameCorpus]:
        """
        :return: the last DataFrameCorpus object that was built. If none have been built, returns None.
        :rtype: Optional[DataFrameCorpus]
        """
        return self.controller.get_latest_corpus()

    def get_corpus(self, corpus_name: str) -> Optional[DataFrameCorpus]:
        """
        :return: the DataFrameCorpus corresponding to the provided name. If no corpus with the given name is found, return None
        :rtype: Optional[DataFrameCorpus]
        """
        return self.controller.get_corpus(corpus_name)

    def get_corpora(self) -> dict[str, DataFrameCorpus]:
        """
        :return: a dictionary that maps Corpus names to DataFrameCorpus objects that have been built using this CorpusLoader
        :rtype: dict[str, DataFrameCorpus]
        """
        return self.controller.get_corpora()



import logging
from datetime import datetime
from logging.handlers import RotatingFileHandler
from io import BytesIO
from os.path import abspath, join, dirname
from typing import Optional, Callable

from atap_corpus.corpus.corpus import DataFrameCorpus
from pandas import DataFrame
from panel.widgets import Tqdm

from atap_corpus_loader.controller.CorpusExportService import CorpusExportService
from atap_corpus_loader.controller.FileLoaderService import FileLoaderService, FileLoadError
from atap_corpus_loader.controller.OniAPIService import OniAPIService
from atap_corpus_loader.controller.data_objects import (FileReference, ViewCorpusInfo, CorpusHeader, DataType,
                                                        UniqueNameCorpora)
from atap_corpus_loader.controller.file_loader_strategy.FileLoaderFactory import ValidFileType
from atap_corpus_loader.view.notifications import NotifierService


class Controller:
    LOGGER: logging.Logger = None
    """
    Provides methods for indirection between the corpus loading logic and the user interface
    Holds a reference to the latest corpus built.
    The build_callback_fn will be called when a corpus is built (can be set using set_build_callback()).
    """
    def __init__(self, root_directory: str):
        Controller.setup_logger()

        self.file_loader_service: FileLoaderService = FileLoaderService(root_directory)
        self.oni_api_service: OniAPIService = OniAPIService()
        self.corpus_export_service: CorpusExportService = CorpusExportService()
        self.notifier_service: NotifierService = NotifierService()

        self.text_header: Optional[CorpusHeader] = None
        self.corpus_link_header: Optional[CorpusHeader] = None
        self.meta_link_header: Optional[CorpusHeader] = None

        self.corpus_headers: list[CorpusHeader] = []
        self.meta_headers: list[CorpusHeader] = []

        self.corpora: UniqueNameCorpora = UniqueNameCorpora()

        self.build_callback_fn: Optional[Callable] = None
        self.build_callback_args: list = []
        self.build_callback_kwargs: dict = {}

        self.tqdm_obj = Tqdm()
        self.tqdm_obj.visible = False

    @staticmethod
    def setup_logger():
        if Controller.LOGGER is not None:
            return
        formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        log_file_location = abspath(join(dirname(__file__), '..', 'log.txt'))
        # Max size is ~10MB with 1 backup, so a max size of ~20MB for log files
        max_bytes: int = 10000000
        backup_count: int = 1
        file_handler = RotatingFileHandler(log_file_location, maxBytes=max_bytes, backupCount=backup_count)
        file_handler.setLevel(logging.DEBUG)
        file_handler.setFormatter(formatter)

        console_handler = logging.StreamHandler()
        console_handler.setLevel(logging.ERROR)
        console_handler.setFormatter(formatter)

        Controller.LOGGER = logging.getLogger(__name__)
        Controller.LOGGER.setLevel(logging.DEBUG)
        Controller.LOGGER.addHandler(file_handler)
        Controller.LOGGER.addHandler(console_handler)

        Controller.LOGGER.info('Logger started')

    def display_error(self, error_msg: str):
        Controller.LOGGER.error(f"Error displayed: {error_msg}")
        self.notifier_service.notify_error(error_msg)

    def display_success(self, success_msg: str):
        Controller.LOGGER.info(f"Success displayed: {success_msg}")
        self.notifier_service.notify_success(success_msg)

    def set_build_callback(self, callback: Callable, *args, **kwargs):
        if not callable(callback):
            raise ValueError("Provided callback function must be a callable")
        self.build_callback_fn = callback
        self.build_callback_args = args
        self.build_callback_kwargs = kwargs

    def get_latest_corpus(self) -> Optional[DataFrameCorpus]:
        if len(self.corpora) == 0:
            return
        return self.corpora.items()[-1]

    def get_corpus(self, corpus_name: str) -> Optional[DataFrameCorpus]:
        return self.corpora.get(corpus_name)

    def get_corpora(self) -> dict[str, DataFrameCorpus]:
        corpora_list: list = self.corpora.items()
        corpora_dict: dict[str, DataFrameCorpus] = {}
        for corpus in corpora_list:
            corpora_dict[corpus.name] = corpus

        return corpora_dict

    def load_corpus_from_filepaths(self, filepath_ls: list[str], include_hidden: bool) -> bool:
        Controller.LOGGER.debug(f"Files loaded as corpus: {filepath_ls}")
        try:
            self.file_loader_service.add_corpus_files(filepath_ls, include_hidden, self.tqdm_obj)
            self.corpus_headers = self.file_loader_service.get_inferred_corpus_headers()
        except FileLoadError as e:
            self.display_error(str(e))
            self.unload_filepaths(filepath_ls)
            return False

        return True


    def load_meta_from_filepaths(self, filepath_ls: list[str], include_hidden: bool) -> bool:
        Controller.LOGGER.debug(f"Files loaded as meta: {filepath_ls}")
        try:
            self.file_loader_service.add_meta_files(filepath_ls, include_hidden, self.tqdm_obj)
            self.meta_headers = self.file_loader_service.get_inferred_meta_headers()
        except FileLoadError as e:
            self.display_error(str(e))
            self.unload_filepaths(filepath_ls)
            return False

        return True

    def build_corpus(self, corpus_name: str) -> bool:
        if self.is_meta_added():
            if (self.corpus_link_header is None) or (self.meta_link_header is None):
                self.display_error("Cannot build without link headers set. Select a corpus header and a meta header as linking headers in the dropdowns")
                return False

        if (corpus_name == '') or (corpus_name is None):
            corpus_name = f"Corpus-{datetime.now()}"

        self.tqdm_obj.visible = True
        try:
            corpus = self.file_loader_service.build_corpus(corpus_name, self.corpus_headers,
                                                           self.meta_headers, self.text_header,
                                                           self.corpus_link_header, self.meta_link_header,
                                                           self.tqdm_obj)
        except FileLoadError as e:
            Controller.LOGGER.exception("Exception while building corpus: ")
            self.display_error(str(e))
            self.tqdm_obj.visible = False
            return False
        except Exception as e:
            Controller.LOGGER.exception("Exception while building corpus: ")
            self.display_error(f"Unexpected error building corpus: {e}")
            self.tqdm_obj.visible = False
            return False

        try:
            self.corpora.add(corpus)
        except Exception as e:
            Controller.LOGGER.exception("Exception while adding corpus to corpora: ")
            self.display_error(str(e))
            self.tqdm_obj.visible = False
            return False

        try:
            if self.build_callback_fn is not None:
                self.build_callback_fn(*self.build_callback_args, **self.build_callback_kwargs)
        except Exception as e:
            Controller.LOGGER.exception("Exception while calling build callback: ")
            self.display_error(f"Build callback error: {e}")
            self.tqdm_obj.visible = False
            return False

        self.tqdm_obj.visible = False

        return True

    def get_corpora_info(self) -> list[ViewCorpusInfo]:
        corpora_info: list[ViewCorpusInfo] = []

        for corpus in self.corpora.items():
            corpus_as_df: DataFrame = corpus.to_dataframe()

            name: Optional[str] = corpus.name
            num_rows: int = len(corpus)
            headers: list[str] = []
            dtypes: list[str] = []
            dtype: str
            for header_name, dtype_obj in corpus_as_df.dtypes.items():
                try:
                    dtype = DataType(str(dtype_obj).lower()).name
                except ValueError:
                    dtype = DataType.TEXT.name
                dtypes.append(dtype)
                headers.append(str(header_name))

            corpora_info.append(ViewCorpusInfo(name, num_rows, headers, dtypes))

        return corpora_info

    def delete_corpus(self, corpus_name: str):
        self.corpora.remove(corpus_name)

    def rename_corpus(self, corpus_name: str, new_name: str):
        corpus: Optional[DataFrameCorpus] = self.corpora.get(corpus_name)
        if corpus is None:
            self.display_error(f"No corpus with name {corpus_name} found")
            return

        try:
            corpus.rename(new_name)
        except ValueError as e:
            self.display_error(str(e))

    def get_loaded_file_counts(self) -> dict[str, int]:
        corpus_file_set = set(self.file_loader_service.get_loaded_corpus_files())
        meta_file_set = set(self.file_loader_service.get_loaded_meta_files())
        file_set = corpus_file_set | meta_file_set

        file_counts: dict[str, int] = {"Total files": len(file_set)}
        for file_ref in file_set:
            extension = file_ref.get_extension().upper()
            if file_counts.get(extension) is None:
                file_counts[extension] = 1
            else:
                file_counts[extension] += 1

        return file_counts

    def unload_filepaths(self, filepath_ls: list[str]):
        for filepath in filepath_ls:
            self.file_loader_service.remove_meta_filepath(filepath)
            self.file_loader_service.remove_corpus_filepath(filepath)

        if len(self.file_loader_service.get_loaded_corpus_files()) == 0:
            self.text_header = None
            self.corpus_headers = []
            self.corpus_link_header = None
        if len(self.file_loader_service.get_loaded_meta_files()) == 0:
            self.meta_headers = []
            self.meta_link_header = None

    def unload_all(self):
        Controller.LOGGER.debug("All files unloaded")
        self.file_loader_service.remove_all_files()

        self.text_header = None
        self.corpus_headers = []
        self.meta_headers = []
        self.corpus_link_header = None
        self.meta_link_header = None

    def get_loaded_corpus_files(self) -> set[FileReference]:
        return self.file_loader_service.get_loaded_corpus_files_set()

    def get_loaded_meta_files(self) -> set[FileReference]:
        return self.file_loader_service.get_loaded_meta_files_set()

    def get_corpus_headers(self) -> list[CorpusHeader]:
        return self.corpus_headers

    def get_meta_headers(self) -> list[CorpusHeader]:
        return self.meta_headers

    def get_text_header(self) -> Optional[CorpusHeader]:
        return self.text_header

    def get_corpus_link_header(self) -> Optional[CorpusHeader]:
        return self.corpus_link_header

    def get_meta_link_header(self) -> Optional[CorpusHeader]:
        return self.meta_link_header

    def get_all_datatypes(self) -> list[str]:
        return [d.name for d in DataType]

    def get_valid_filetypes(self) -> list[str]:
        return [ft.name for ft in ValidFileType]

    def is_corpus_added(self) -> bool:
        return len(self.file_loader_service.get_loaded_corpus_files()) > 0

    def is_meta_added(self) -> bool:
        return len(self.file_loader_service.get_loaded_meta_files()) > 0

    def update_corpus_header(self, header: CorpusHeader, include: Optional[bool], datatype_name: Optional[str]):
        if include is not None:
            header.include = include
        if datatype_name is not None:
            header.datatype = DataType[datatype_name]

        for i, corpus_header in enumerate(self.corpus_headers):
            if header == corpus_header:
                self.corpus_headers[i] = header

    def update_meta_header(self, header: CorpusHeader, include: Optional[bool], datatype_name: Optional[str]):
        if include is not None:
            header.include = include
        if datatype_name is not None:
            header.datatype = DataType[datatype_name]

        for i, meta_header in enumerate(self.meta_headers):
            if header == meta_header:
                self.meta_headers[i] = header

    def set_text_header(self, text_header: Optional[str]):
        if text_header is None:
            self.text_header = None
            return

        for header in self.corpus_headers:
            if header.name == text_header:
                self.text_header = header
                header.datatype = DataType.TEXT
                header.include = True
                return

    def set_corpus_link_header(self, link_header_name: Optional[str]):
        for header in self.corpus_headers:
            if header.name == link_header_name:
                self.corpus_link_header = header
                header.include = True
                return
        self.corpus_link_header = None

    def set_meta_link_header(self, link_header_name: Optional[str]):
        for header in self.meta_headers:
            if header.name == link_header_name:
                self.meta_link_header = header
                header.include = True
                return
        self.meta_link_header = None

    def retrieve_all_files(self, expand_archived: bool) -> list[FileReference]:
        return self.file_loader_service.get_all_files(expand_archived)

    def get_export_types(self) -> list[str]:
        return self.corpus_export_service.get_filetypes()

    def export_corpus(self, corpus_name: str, filetype: str) -> Optional[BytesIO]:
        corpus: Optional[DataFrameCorpus] = self.corpora.get(corpus_name)
        if corpus is None:
            self.display_error(f"No corpus with name '{corpus_name}' found")
            return None

        try:
            return self.corpus_export_service.export(corpus.to_dataframe(), filetype)
        except ValueError as e:
            self.display_error(str(e))



from io import BytesIO
from typing import Callable

from pandas import DataFrame


class CorpusExportService:
    def __init__(self):
        self.export_type_mapping: dict[str, Callable] = {
            'csv': self.export_csv,
            'xlsx': self.export_xlsx
        }

    def get_filetypes(self) -> list[str]:
        return list(self.export_type_mapping.keys())

    def export(self, df: DataFrame, filetype: str) -> BytesIO:
        if filetype not in self.export_type_mapping:
            raise ValueError(f"{filetype} is not a valid export format")
        file_object: BytesIO = self.export_type_mapping[filetype](df)
        file_object.seek(0)

        return file_object

    @staticmethod
    def export_csv(df: DataFrame) -> BytesIO:
        csv_object = BytesIO()
        df.to_csv(csv_object, index=False)

        return csv_object

    @staticmethod
    def export_xlsx(df: DataFrame) -> BytesIO:
        excel_object = BytesIO()
        df.to_excel(excel_object, index=False)

        return excel_object



from panel import Tabs

from atap_corpus_loader.controller import Controller
from atap_corpus_loader.view.gui import AbstractWidget, FileLoaderWidget, CorpusInfoWidget


class ViewWrapperWidget(AbstractWidget):
    """
    A wrapper class that holds different loading method interfaces within a Tab
    """
    def __init__(self, controller: Controller):
        super().__init__()
        self.controller: Controller = controller

        self.file_loader: FileLoaderWidget = FileLoaderWidget(self, controller)
        self.corpus_display: CorpusInfoWidget = CorpusInfoWidget(controller)

        self.panel = Tabs(("File Loader", self.file_loader),
                          ("Corpus Overview", self.corpus_display))
        self.corpus_info_idx: int = len(self.panel) - 1
        self.children = [self.file_loader, self.corpus_display]

    def update_display(self):
        pass

    def load_corpus_from_filepaths(self, filepath_ls: list[str], include_hidden: bool):
        if len(filepath_ls) == 0:
            return
        success = self.controller.load_corpus_from_filepaths(filepath_ls, include_hidden)
        self.update_displays()
        if success:
            self.controller.display_success("Corpus files loaded successfully")

    def load_meta_from_filepaths(self, filepath_ls: list[str], include_hidden: bool):
        if len(filepath_ls) == 0:
            return
        success = self.controller.load_meta_from_filepaths(filepath_ls, include_hidden)
        self.update_displays()
        if success:
            self.controller.display_success("Metadata files loaded successfully")

    def build_corpus(self, corpus_name: str) -> bool:
        success: bool = self.controller.build_corpus(corpus_name)
        if success:
            self.update_displays()

            self.panel.active = self.corpus_info_idx
            corpus_name: str = self.controller.get_latest_corpus().name
            self.controller.display_success(f"Corpus {corpus_name} built successfully")

        return success



from panel import Row

from atap_corpus_loader.controller import Controller
from atap_corpus_loader.view.gui import AbstractWidget


class OniLoaderWidget(AbstractWidget):
    def __init__(self, controller: Controller):
        super().__init__()
        self.controller: Controller = controller

        self.panel = Row()

    def update_display(self):
        pass



from io import BytesIO
from typing import Optional

import panel
from panel import Row, Accordion, bind, HSpacer
from panel.pane import Markdown
from panel.widgets import Button, TextInput, FileDownload, Select

from atap_corpus_loader.controller import Controller
from atap_corpus_loader.controller.data_objects import ViewCorpusInfo
from atap_corpus_loader.view.gui import AbstractWidget

panel.extension()


class CorpusInfoWidget(AbstractWidget):
    def __init__(self, controller: Controller):
        super().__init__()
        self.controller: Controller = controller

        self.corpus_controls = Accordion(toggle=True, width=600)
        self.corpus_controls.header_background = "#FFFFFA"
        self.corpus_controls.active_header_background = "#7A8B99"
        self.corpus_controls.param.watch(self._update_corpus_display, 'active', onlychanged=False)

        self.corpus_display: Markdown = Markdown()

        self.panel = Row(self.corpus_controls, self.corpus_display)

    @staticmethod
    def _build_corpus_label(corpus_info: ViewCorpusInfo) -> str:
        name: Optional[str] = corpus_info.name
        if name is None:
            name = " "
        row_info: str = f"{corpus_info.num_rows} document"
        if corpus_info.num_rows != 1:
            row_info += 's'

        return f"{name} - {row_info}"

    @staticmethod
    def _build_header_markdown_table(corpus_info: ViewCorpusInfo) -> str:
        if len(corpus_info.headers) != len(corpus_info.dtypes):
            return " "

        header_row = "| " + " | ".join(corpus_info.headers) + " |"
        spacer_row = "| :-: " * len(corpus_info.headers) + "|"
        data_row = "| " + " | ".join(corpus_info.dtypes) + " |"

        header_table_text = f"{header_row}\n{spacer_row}\n{data_row}"
        return header_table_text

    def export_corpus(self, corpus_name: str, filetype: str) -> Optional[BytesIO]:
        return self.controller.export_corpus(corpus_name, filetype)

    def rename_corpus(self, corpus_name: str, name: str):
        self.controller.rename_corpus(corpus_name, name)
        self.update_display()

    def delete_corpus(self, corpus_name: str):
        self.controller.delete_corpus(corpus_name)
        self.update_display()

    def _update_corpus_display(self, *event):
        current_active: list[int] = self.corpus_controls.active
        if len(current_active) == 0:
            self.corpus_display.object = ' '
            self.corpus_display.visible = False
            return

        corpora_info: list[ViewCorpusInfo] = self.controller.get_corpora_info()
        active_idx: int = current_active[0]
        if active_idx >= len(corpora_info):
            self.corpus_display.object = ' '
            self.corpus_display.visible = False
            return

        corpus_info: ViewCorpusInfo = corpora_info[active_idx]
        header_markdown_table: str = CorpusInfoWidget._build_header_markdown_table(corpus_info)

        self.corpus_display.object = header_markdown_table
        self.corpus_display.visible = True

    def update_display(self):
        corpora_info: list[ViewCorpusInfo] = self.controller.get_corpora_info()
        export_types: list[str] = self.controller.get_export_types()
        default_filetype: str = export_types[0]

        corpus_controls_objs: list[Row] = []
        for corpus_info in corpora_info:
            label: str = CorpusInfoWidget._build_corpus_label(corpus_info=corpus_info)

            filetype_dropdown = Select(name="Export filetype", options=export_types, value=default_filetype,
                                       width=100, align="center")
            corpus_export_button = FileDownload(
                label=f"Export",
                filename=f"{corpus_info.name}.{default_filetype}",
                callback=bind(self.export_corpus, corpus_name=corpus_info.name, filetype=filetype_dropdown),
                button_type="primary", button_style="solid",
                height=30, width=100,
                align="center")

            def select_fn(event, name=corpus_info.name, export_button=corpus_export_button):
                filename: str = f"{name}.{event.new}"
                export_button.filename = filename
            filetype_dropdown.param.watch(select_fn, ['value'])

            rename_field: TextInput = TextInput(name="Rename corpus", value=corpus_info.name,
                                                align='center', width=150)
            rename_field.param.watch(lambda event, corpus_name=corpus_info.name: self.rename_corpus(corpus_name, event.new), ['value'])
            delete_button: Button = Button(name="Delete corpus", button_type="danger", align='center')
            delete_button.on_click(lambda event, corpus_name=corpus_info.name: self.delete_corpus(corpus_name))

            corpus_control_row = Row(rename_field, filetype_dropdown, corpus_export_button,
                                     HSpacer(), delete_button, name=label)
            corpus_controls_objs.append(corpus_control_row)

        self.corpus_controls.objects = []
        self.corpus_controls.objects = corpus_controls_objs
        self._update_corpus_display()



from enum import Enum, auto

from atap_corpus_loader.controller.data_objects import FileReference
from atap_corpus_loader.controller.file_loader_strategy.FileLoadError import FileLoadError
from atap_corpus_loader.controller.file_loader_strategy.concrete_strategies import *
from atap_corpus_loader.controller.file_loader_strategy.FileLoaderStrategy import FileLoaderStrategy


class ValidFileType(Enum):
    """
    An enum of valid file extensions that are supported with a loader
    """
    TXT = auto()
    ODT = auto()
    DOCX = auto()
    CSV = auto()
    TSV = auto()
    XLSX = auto()
    ODS = auto()
    RDS = auto()
    RDATA = auto()
    RDA = auto()
    ZIP = auto()


class FileLoaderFactory:
    """
    Provides a single public method to map a FileReference object to concrete FileLoaderStrategy object

    FILETYPE_LOADER_MAP maps file extensions to concrete FileLoaderStrategy types.
    The file extensions are treated as case-insensitive
    """
    FILETYPE_LOADER_MAP: dict = {
        ValidFileType.TXT: TXTLoaderStrategy,
        ValidFileType.ODT: ODTLoaderStrategy,
        ValidFileType.DOCX: DOCXLoaderStrategy,
        ValidFileType.CSV: CSVLoaderStrategy,
        ValidFileType.TSV: TSVLoaderStrategy,
        ValidFileType.XLSX: XLSXLoaderStrategy,
        ValidFileType.ODS: ODSLoaderStrategy,
        ValidFileType.RDS: RLoaderStrategy,
        ValidFileType.RDATA: RLoaderStrategy,
        ValidFileType.RDA: RLoaderStrategy
    }

    @staticmethod
    def get_file_loader(file_ref: FileReference) -> FileLoaderStrategy:
        """
        Maps the provided FileReference object to a concrete FileLoaderStrategy object based on the extension.
        If the file extension is missing (the filename is not of the format <name>.<extension> or is not
        valid, i.e. is not a member of the ValidFileType enum) a FileLoadError will be raised.
        :param file_ref: the FileReference object corresponding to the file to assign a loader to
        :return: a concrete FileLoaderStrategy object that has been passed the provided FileReference object.
        :raises FileLoadError: if there is no '.' in the file name or the extension after the '.' is not a valid file type
        """
        file_name: str = file_ref.get_filename()
        file_extension: str = file_ref.get_extension().upper()
        if file_extension == '':
            raise FileLoadError(f"No file extension found in file name: {file_name}. "
                                f"File name must be in format <filename>.<extension>")

        try:
            file_type: ValidFileType = ValidFileType[file_extension]
            file_loader: FileLoaderStrategy = FileLoaderFactory.FILETYPE_LOADER_MAP[file_type](file_ref)
        except KeyError:
            accepted_types: str = ', '.join([ft.name for ft in ValidFileType])
            raise FileLoadError(f"Invalid file type loaded: {file_extension}. Valid file types: {accepted_types}")

        return file_loader



from enum import Enum


class DataType(Enum):
    """
    Maps readable data type names to the pandas data types
    """
    TEXT = 'string'
    INTEGER = 'int64'
    DECIMAL = 'float64'
    BOOLEAN = 'boolean'
    DATETIME = 'datetime64[ns]'
    CATEGORY = 'category'



from pandas import DataFrame, concat
from pyreadr import list_objects, read_r
from pyreadr.librdata import PyreadrError, LibrdataError

from atap_corpus_loader.controller.data_objects import CorpusHeader, DataType
from atap_corpus_loader.controller.file_loader_strategy.FileLoadError import FileLoadError
from atap_corpus_loader.controller.file_loader_strategy.FileLoaderStrategy import FileLoaderStrategy


class RLoaderStrategy(FileLoaderStrategy):
    def get_inferred_headers(self) -> list[CorpusHeader]:
        real_path: str = self.file_ref.resolve_real_file_path()

        try:
            file_objects: list[dict] = list_objects(real_path)
        except (PyreadrError, LibrdataError) as e:
            raise FileLoadError(f"Error loading R file: {str(e)}")

        if len(file_objects) == 0:
            return []

        columns = file_objects[0].get('columns')
        for file_object in file_objects[1:]:
            if file_object.get('columns') != columns:
                raise FileLoadError(f"Incompatible headers within loaded R objects")

        if len(columns) == 0:
            raise FileLoadError(f"No tabular data found. Ensure the file contains tabular data (rows and columns)")

        headers: list[CorpusHeader] = []
        dtype: DataType = DataType.TEXT
        for col_name in columns:
            headers.append(CorpusHeader(col_name, dtype, True))

        return headers

    def get_dataframe(self, headers: list[CorpusHeader]) -> DataFrame:
        real_path: str = self.file_ref.resolve_real_file_path()

        try:
            object_df_dict: dict = read_r(real_path)
        except (PyreadrError, LibrdataError) as e:
            raise FileLoadError(f"Error loading R file: {str(e)}")

        df_list = object_df_dict.values()
        concat_df: DataFrame = concat(df_list, ignore_index=True)
        excluded_headers: list[str] = [header.name for header in headers if not header.include]
        df = concat_df.drop(excluded_headers, axis='columns')
        dtypes_applied_df: DataFrame = FileLoaderStrategy._apply_selected_dtypes(df, headers)

        return dtypes_applied_df



from glob import iglob
from os import R_OK, access
from os.path import normpath, sep, isdir, exists
from typing import Optional, Iterator

from pandas import DataFrame, merge, concat
from panel.widgets import Tqdm
from atap_corpus.corpus.corpus import DataFrameCorpus

from atap_corpus_loader.controller.data_objects import FileReference, CorpusHeader, FileReferenceFactory
from atap_corpus_loader.controller.file_loader_strategy import FileLoaderStrategy, FileLoaderFactory, FileLoadError

"""
Some methods in this module utilise Tqdm from the panel library, which breaks the Model-View separation.
This has been done out of necessity for a progress bar for particular operations.
The panel Tqdm is a wrapper for the standard tqdm module and can be replaced if needed.
"""


class FileLoaderService:
    """
    Provides methods that handle the logic of loading files and building the DataFrameCorpus object from the loaded
    files.
    Maintains a reference to files loaded as corpus files and files loaded as metadata files.
    """
    def __init__(self, root_directory: str):
        self.root_directory: str = self._sanitise_root_dir(root_directory)
        self.loaded_corpus_files: set[FileReference] = set()
        self.loaded_meta_files: set[FileReference] = set()
        # Utilise FileReferenceFactory.clear_cache() if memory overhead is raised as an issue.
        self.file_ref_factory: FileReferenceFactory = FileReferenceFactory()

        self.all_files_cache: list[FileReference] = []
        self.all_files_count: int = 0

    def get_all_files(self, expand_archived: bool) -> list[FileReference]:
        path_iter: Iterator = iglob(f"{self.root_directory}**", recursive=True)
        all_file_refs: list[FileReference] = []
        for path in path_iter:
            if isdir(path):
                continue

            file_refs: list[FileReference] = self.file_ref_factory.get_file_refs_from_path(path, expand_archived)
            all_file_refs.extend(file_refs)

        all_file_refs.sort(key=lambda ref: ref.get_path())

        return all_file_refs

    def get_loaded_corpus_files(self) -> list[FileReference]:
        return [f for f in self.loaded_corpus_files if not f.is_archive()]

    def get_loaded_meta_files(self) -> list[FileReference]:
        return [f for f in self.loaded_meta_files if not f.is_archive()]

    def get_loaded_corpus_files_set(self) -> set[FileReference]:
        return set(self.get_loaded_corpus_files())

    def get_loaded_meta_files_set(self) -> set[FileReference]:
        return set(self.get_loaded_meta_files())

    def add_corpus_files(self, corpus_filepaths: list[str], include_hidden: bool, tqdm_obj: Tqdm):
        for filepath in tqdm_obj(corpus_filepaths, desc="Loading corpus files", unit="files", leave=False):
            file_ref: FileReference = self.file_ref_factory.get_file_ref(filepath)
            if file_ref in self.loaded_corpus_files:
                continue
            if not include_hidden and file_ref.is_hidden():
                continue
            FileLoaderService._check_filepath_permissions(file_ref)

            self.loaded_corpus_files.add(file_ref)
            if file_ref.is_archive():
                zip_refs: list[FileReference] = self.file_ref_factory.get_zip_file_refs(filepath)
                for zip_ref in zip_refs:
                    if not zip_ref.is_hidden() or include_hidden:
                        self.loaded_corpus_files.add(zip_ref)

    def add_meta_files(self, meta_filepaths: list[str], include_hidden: bool, tqdm_obj: Tqdm):
        for filepath in tqdm_obj(meta_filepaths, desc="Loading metadata files", unit="files", leave=False):
            file_ref: FileReference = self.file_ref_factory.get_file_ref(filepath)
            if file_ref in self.loaded_meta_files:
                continue
            if not include_hidden and file_ref.is_hidden():
                continue
            FileLoaderService._check_filepath_permissions(file_ref)

            self.loaded_meta_files.add(file_ref)
            if file_ref.is_archive():
                zip_refs: list[FileReference] = self.file_ref_factory.get_zip_file_refs(filepath)
                for zip_ref in zip_refs:
                    if not zip_ref.is_hidden() or include_hidden:
                        self.loaded_meta_files.add(zip_ref)

    def remove_corpus_filepath(self, corpus_filepath: str):
        file_ref: FileReference = self.file_ref_factory.get_file_ref(corpus_filepath)
        if file_ref in self.loaded_corpus_files:
            self.loaded_corpus_files.remove(file_ref)

    def remove_meta_filepath(self, meta_filepath: str):
        file_ref: FileReference = self.file_ref_factory.get_file_ref(meta_filepath)
        if file_ref in self.loaded_meta_files:
            self.loaded_meta_files.remove(file_ref)

    def remove_loaded_corpus_files(self):
        self.loaded_corpus_files.clear()

    def remove_loaded_meta_files(self):
        self.loaded_meta_files.clear()

    def remove_all_files(self):
        self.remove_loaded_corpus_files()
        self.remove_loaded_meta_files()

    def get_inferred_corpus_headers(self) -> list[CorpusHeader]:
        return self._get_file_headers(self.get_loaded_corpus_files())

    def get_inferred_meta_headers(self) -> list[CorpusHeader]:
        return self._get_file_headers(self.get_loaded_meta_files())

    def _get_file_headers(self, file_refs: list[FileReference]) -> list[CorpusHeader]:
        headers: Optional[list[CorpusHeader]] = None
        for ref in file_refs:
            file_loader: FileLoaderStrategy = FileLoaderFactory.get_file_loader(ref)
            try:
                path_headers: list[CorpusHeader] = file_loader.get_inferred_headers()
            except UnicodeDecodeError:
                self.remove_corpus_filepath(ref.get_path())
                self.remove_meta_filepath(ref.get_path())
                raise FileLoadError(f"Error loading file at {ref.get_path()}: file is not UTF-8 encoded")
            except Exception as e:
                self.remove_corpus_filepath(ref.get_path())
                self.remove_meta_filepath(ref.get_path())
                raise FileLoadError(f"Error loading file at {ref.get_path()}: {e}")

            if headers is None:
                headers = path_headers
            elif set(headers) != set(path_headers):
                self.remove_corpus_filepath(ref.get_path())
                self.remove_meta_filepath(ref.get_path())
                raise FileLoadError(f"Incompatible data labels in file: {ref.get_path()}")

        if headers is None:
            headers = []

        return headers

    def build_corpus(self, corpus_name: str,
                     corpus_headers: list[CorpusHeader],
                     meta_headers: list[CorpusHeader],
                     text_header: CorpusHeader,
                     corpus_link_header: Optional[CorpusHeader],
                     meta_link_header: Optional[CorpusHeader],
                     tqdm_obj: Tqdm) -> DataFrameCorpus:
        corpus_df: DataFrame = FileLoaderService._get_concatenated_dataframe(self.get_loaded_corpus_files(),
                                                                             corpus_headers,
                                                                             tqdm_obj,
                                                                             "Building corpus")
        meta_df: DataFrame = FileLoaderService._get_concatenated_dataframe(self.get_loaded_meta_files(),
                                                                           meta_headers,
                                                                           tqdm_obj,
                                                                           "Building metadata")

        load_corpus: bool = len(corpus_headers) > 0
        load_meta: bool = len(meta_headers) > 0

        final_df: DataFrame
        if load_corpus and load_meta:
            final_df = merge(left=corpus_df, right=meta_df,
                             left_on=corpus_link_header.name, right_on=meta_link_header.name,
                             how='inner', suffixes=(None, '_meta'))
        elif load_corpus:
            final_df = corpus_df
        elif load_meta:
            final_df = meta_df
        else:
            raise ValueError("No corpus headers or metadata headers provided")

        return DataFrameCorpus.from_dataframe(final_df, text_header.name, corpus_name)

    @staticmethod
    def _sanitise_root_dir(root_directory: str) -> str:
        if type(root_directory) is not str:
            raise TypeError(f"root_directory argument: expected string, got {type(root_directory)}")
        sanitised_directory = normpath(root_directory)

        if not sanitised_directory.endswith(sep):
            sanitised_directory += sep

        return sanitised_directory

    @staticmethod
    def _check_filepath_permissions(file_ref: FileReference):
        filepath: str
        if file_ref.is_zipped():
            filepath = file_ref.get_directory_path()
        else:
            filepath = file_ref.get_path()
        if not exists(filepath):
            raise FileLoadError(f"No file found at: {filepath}")
        if not access(filepath, R_OK):
            raise FileLoadError(f"No permissions to read the file at: {filepath}")

    @staticmethod
    def _get_concatenated_dataframe(file_refs: list[FileReference],
                                    headers: list[CorpusHeader],
                                    tqdm_obj: Tqdm,
                                    loading_msg: str) -> DataFrame:
        if len(file_refs) == 0:
            return DataFrame()
        df_list: list[DataFrame] = []
        for ref in tqdm_obj(file_refs, desc=loading_msg, unit="files", leave=False):
            file_loader: FileLoaderStrategy = FileLoaderFactory.get_file_loader(ref)
            try:
                path_df: DataFrame = file_loader.get_dataframe(headers)
            except UnicodeDecodeError:
                raise FileLoadError(f"Error loading file at {ref.get_path()}: file is not UTF-8 encoded")
            except Exception as e:
                raise FileLoadError(f"Error loading file at {ref.get_path()}: {e}")

            df_list.append(path_df)

        return concat(df_list, ignore_index=True)



from fnmatch import fnmatch
from typing import Callable

import panel
from panel import Row, Column
from panel.widgets import Button, MultiSelect, TextInput, Select, Checkbox

from atap_corpus_loader.controller import Controller
from atap_corpus_loader.controller.data_objects import FileReference
from atap_corpus_loader.view.gui import AbstractWidget


class FileSelectorWidget(AbstractWidget):
    def __init__(self, view_handler: AbstractWidget, controller: Controller):
        super().__init__()
        self.view_handler: AbstractWidget = view_handler
        self.controller: Controller = controller
        # The function will be set by an instance method, but must be callable until then
        self._set_button_status_on_operation: Callable = lambda curr_loading: None

        self.select_all_button = Button(name="Select all", width=95,
                                        button_style="solid", button_type="primary")
        self.select_all_button.on_click(self.select_all)

        self.filter_input = TextInput(placeholder="Filter displayed files (supports wildcard syntax)",
                                      sizing_mode='stretch_width')
        self.filter_input.param.watch(self._on_filter_change, ['value'])

        self.show_hidden_files_checkbox = Checkbox(name="Show hidden", value=False, align="start")
        self.show_hidden_files_checkbox.param.watch(self._on_filter_change, ['value'])

        self.expand_archive_checkbox = Checkbox(name="Expand archives", value=False, align="start")
        self.expand_archive_checkbox.param.watch(self._on_filter_change, ['value'])

        self.file_type_filter = Select(name='Filter by filetype', width=150)
        self.file_type_filter.options = ['All valid filetypes'] + self.controller.get_valid_filetypes()
        self.file_type_filter.param.watch(self._on_filter_change, ['value'])

        self.selector_widget = MultiSelect(size=20, sizing_mode='stretch_width')

        self.panel = Column(
            Row(self.select_all_button),
            Row(self.filter_input,
                Column(
                    self.show_hidden_files_checkbox,
                    self.expand_archive_checkbox
                ),
                self.file_type_filter),
            Row(self.selector_widget),
            width=700)

        panel.state.add_periodic_callback(self.update_display, period=2000)
        self.update_display()

    def set_button_operation_fn(self, _set_button_status_on_operation: Callable):
        self._set_button_status_on_operation = _set_button_status_on_operation

    def update_display(self):
        loaded_corpus_files: set[FileReference] = self.controller.get_loaded_corpus_files()
        loaded_meta_files: set[FileReference] = self.controller.get_loaded_meta_files()

        filtered_refs: list[FileReference] = self._get_filtered_file_refs()

        filtered_files_dict: dict[str, str] = {}
        checkmark_symbol = "\U00002714"
        for ref in filtered_refs:
            file_repr = ref.get_path()
            if ref in loaded_corpus_files:
                file_repr += f" {checkmark_symbol} [corpus]"
            if ref in loaded_meta_files:
                file_repr += f" {checkmark_symbol} [meta]"
            filtered_files_dict[file_repr] = ref.get_path()

        self.selector_widget.options = filtered_files_dict

    def _get_filtered_file_refs(self) -> list[FileReference]:
        valid_file_types: list[str] = self.controller.get_valid_filetypes()
        selected_file_types: set[str]
        if self.file_type_filter.value in valid_file_types:
            selected_file_types = {self.file_type_filter.value.upper()}
        else:
            selected_file_types = {ft.upper() for ft in valid_file_types}

        expand_archived: bool = self.expand_archive_checkbox.value
        file_refs: list[FileReference] = self.controller.retrieve_all_files(expand_archived)

        filtered_refs: list[FileReference] = []
        filter_str = f"*{self.filter_input.value}*"
        skip_hidden: bool = not self.show_hidden_files_checkbox.value
        for ref in file_refs:
            if ref.get_extension().upper() not in selected_file_types:
                continue
            if not fnmatch(ref.get_path(), filter_str):
                continue
            if skip_hidden and ref.is_hidden():
                continue

            filtered_refs.append(ref)

        return filtered_refs

    def _on_filter_change(self, *_):
        self._set_button_status_on_operation(curr_loading=True)
        self.update_display()
        self._set_button_status_on_operation(curr_loading=False)

    def select_all(self, *_):
        self._set_button_status_on_operation(curr_loading=True)
        self.selector_widget.value = list(self.selector_widget.options.values())
        self._set_button_status_on_operation(curr_loading=False)

    def get_selector_value(self) -> list[str]:
        return self.selector_widget.value

    def get_show_hidden_value(self) -> bool:
        return self.show_hidden_files_checkbox.value



from typing import Optional


class ViewCorpusInfo:
    def __init__(self, name: Optional[str], num_rows: int, headers: list[str], dtypes: list[str]):
        self.name: Optional[str] = name
        self.num_rows: int = num_rows
        self.headers: list[str] = headers
        self.dtypes: list[str] = dtypes

    def __repr__(self):
        return f"ViewCorpusInfo - name: {self.name}"



from .TXTLoaderStrategy import TXTLoaderStrategy
from .ODTLoaderStrategy import ODTLoaderStrategy
from .DOCXLoaderStrategy import DOCXLoaderStrategy
from .CSVLoaderStrategy import CSVLoaderStrategy
from .TSVLoaderStrategy import TSVLoaderStrategy
from .XLSXLoaderStrategy import XLSXLoaderStrategy
from .ODSLoaderStrategy import ODSLoaderStrategy
from .RLoaderStrategy import RLoaderStrategy
